@@ -26,7 +26,7 @@ use Time::HiRes qw/time/;
 Readonly my $HTTP_OK => 200;
 Readonly my $HTTP_FILE_NOT_FOUND => 404;
 
-our $VERSION = '1.12';
+our $VERSION = '1.13';
 
 # TODO: Timeout, fallback
 # TODO: Expected result content (json etc)
@@ -174,12 +174,18 @@ sub start {
     if ($self->is_cacheable($key)) {
       my $serialized = $self->cache_agent->get($key);
       if ($serialized) {
+        $serialized->{events} = $tx->{events};
+        $serialized->{req_events} = $tx->req->{events};
+        $serialized->{res_events} = $tx->res->{events};
         my $cached_tx = _build_fake_tx($serialized);
         $self->_log_line($cached_tx, {
           start_time => $start_time,
           key => $key,
           type => 'cached result',
         });
+        $cached_tx->req->finish;
+        $cached_tx->res->finish;
+        $cached_tx->closed;
         return $cb->($self, $cached_tx) if $cb;
         return $cached_tx;
       }
@@ -207,7 +213,7 @@ sub start {
       $code = $res->{code};
     } or $self->logger->error($EVAL_ERROR);
 
-    my $params = { url => $url, body => $res->{body}, code => $code, method => 'FILE', headers => $res->{headers} };
+    my $params = { url => $url, body => $res->{body}, code => $code, method => 'FILE', headers => $res->{headers}, events => $tx->{events}, req_events => $tx->req->{events}, res_events => $tx->res->{events} };
 
     # first non-blocking, if no callback, regular post process
     my $tx = _build_fake_tx($params);
@@ -234,6 +240,9 @@ sub _post_process_get {
                 if (my $cache_obj = $self->cache_agent->get_object($key)) {
                     my $serialized = $cache_obj->value;
                     $serialized->{headers}->{'X-Mojo-UserAgent-Cached-ExpiresAt'} = $cache_obj->expires_at($key);
+                    $serialized->{events} = $tx->{events};
+                    $serialized->{req_events} = $tx->req->{events};
+                    $serialized->{res_events} = $tx->res->{events};
 
                     my $expired_tx = _build_fake_tx($serialized);
                     $self->_log_line( $expired_tx, {
@@ -242,6 +251,9 @@ sub _post_process_get {
                         type       => 'expired and cached',
                         orig_tx    => $tx,
                     });
+                    $expired_tx->req->finish;
+                    $expired_tx->res->finish;
+                    $expired_tx->closed;
 
                     return $expired_tx;
                 }
@@ -370,6 +382,10 @@ sub _build_fake_tx {
     $tx->res->{json} = $opts->{json};
     $tx->res->body($opts->{body});
 
+    $tx->{events} = $opts->{events};
+    $tx->req->{events} = $opts->{req_events};
+    $tx->res->{events} = $opts->{res_events};
+
     return $tx;
 }
 
@@ -493,6 +509,8 @@ sub _get_stacktrace {
     } grep { $_ } @frames;
 }
 
+sub _url { shift->req->url->to_abs }
+
 1;
 
 =encoding utf8
